# -*-Tcl-*- (nowrap)
######################################################################
# file:        opetopes2dot.tcl
# author:      Joachim Kock
# last update: 2006-10-17 14:52
######################################################################
# This file defines procedures for producing graphical representation
# of XML encoded opetopes by generating DOT files, and by way of a dot 
# interpreter, also PDF or JPG output, etc.
# 
# The notion of opetope (invented by J.Baez and J.Dolan) and the algorithms
# implemented by the procedures refer to the paper 'Polynomial functors and
# opetopes' by J.Kock, A.Joyal, M.Batanin, and J.-F.Mascari.
# 
# 
# Copyright (c) 2004-2006, Joachim Kock.  All rights reserved.  
# Distributed under a BSD-style license (see bottom of file).


# --------------------------------------------------------------------
# SECTION 0  --  Preliminaries
# --------------------------------------------------------------------

if { [catch {package require tdom}] } {
    if { [info exists argv] } {
	# This file was sourced from a shell script.  Then we use puts instead of error:
	puts "ERROR: Could not load tDOM.  This package is essential, and nothing will work\n \
	  without it.  Please install tDOM.  See the Manual for instructions."
	exit 1
    } else {
	error "Could not load tDOM.  This package is essential, and nothing will work without it.\
	  Please install tDOM.  See the Manual for instructions."
    }
}


# The file opetope.dtd, needed pr parsing the XML files, should reside in
# ../xml/opetope.dtd, compared to the location of this file.  The problem
# is that this file doesn't know its own location.  We can try to figure it
# out: If the variable $homeOfThisScript is set (this is the case if the
# file is sourced from a shell script), then we rely on that one:
if { ![info exists homeOfThisScript] } {
    if { [namespace exists ::alpha] } {
	# If we are running this script from Alpha, then perhaps the file is
	# open in a window, in which case we can figure out where it came from.
	set path [lsearch -glob -inline [winNames -f] *opetopes.tcl]
	if { [string length $path] } {
	    set nameOfThisScript [file normalize $path]
	    set homeOfThisScript [file dirname $nameOfThisScript]
	}
	unset path
    } else {
	# If we still haven't found out, it might be that we are running in tclsh.
	# If this file was sourced through a [source] command, we can figure out
	# its path by looking in the history list!
	set zeroCmd [history event 0]
	puts $zeroCmd
	set i [lsearch -exact $zeroCmd "source"]
	if { $i != -1 } {
	    set filename [lindex $zeroCmd [expr {$i+1}]]
	    if { [file readable $filename] } {
		set nameOfThisScript [file normalize $filename] 
		puts $nameOfThisScript
		set homeOfThisScript [file dirname $nameOfThisScript] 
	    }
	}
    }
}


if { [info exists homeOfThisScript] } {
    set baseURI [file normalize [file join $homeOfThisScript .. xml]]
} else {
    # Let's see how far we get without a baseURI:
    set baseURI ""
    puts "WARNING: Could not figure out the location of this file."
    puts "Please help the script by writing in the beginning of it:"
    puts "    set homeOfThisScript /full/path/to/the/tcl-directory"
}


# Now figure out where dot is.  We read the "configuration" file
# (which happens to be written in sh language, so emulate the
# export statement):
proc export { equation } {
    if { [regexp {^([A-Z]+)=(.*)$} $equation -> VAR VAL] } {
	set ::env($VAR) $VAL
    }
}
source [file join $homeOfThisScript configuration]
if { [catch {set DOT $env(DOTBINARY)}] } {
    set DOT [exec which dot]
}



# A practical key binding for Alpha:
if { [namespace exists ::alpha] } {
    # Make a key binding for XML mode to perform
    # xml-dot-ps-pdf conversion, and open the resulting pdf.
    proc seeXML {} {
	set xmlFile [win::Current]
	# We ought to check that it is an opetope, etc.
	# but we don't care...
	set pdfOut [opetope2pdf $xmlFile]
	#     file delete $psOut
	viewPDFFile $pdfOut
    }
    Bind  't' <c> seeXML xml
}




# --------------------------------------------------------------------
# SECTION 1  --  Write dot file
# --------------------------------------------------------------------

proc opetope2dot { xmlFile } {
    set fd [open $xmlFile r]
    set xml [read $fd]
    close $fd
    
    set doc [dom parse $xml]
    set dotTxt [produceDotCode $doc]
    $doc delete
    
    set fileName [file rootname $xmlFile]
    set dotFileName "$fileName.dot"
    set fd [open $dotFileName w]
    puts $fd $dotTxt
    close $fd
    return $dotFileName
}

proc opetope2ps { xmlFile } {
    set file [file join [pwd] $xmlFile]
    set dotFile [opetope2dot $file]
    set psOut [file rootname $file].ps
    catch {exec $::DOT -Tps $dotFile -o $psOut}
    return $psOut
}

proc opetope2pdf { xmlFile } {
    set file [file join [pwd] $xmlFile]
    set ps [opetope2ps $file]
    set pdfOut [file rootname $ps].pdf
    exec ps2pdf13 $ps $pdfOut
    return $pdfOut
}


proc produceDotCode { doc } {
    
    global homeOfThisScript
    set root [$doc documentElement]
    set rootname [$root @name]
    
    # Now build the dot file:
    set txt "graph \"$rootname\" "
    append txt \{ \n
    append txt "label=\"\"\n"
    
    set fd [open [file normalize [file join $homeOfThisScript .. dot opetope-settings.dot]] r]    
    append txt [read $fd]
    close $fd
    
    append txt \n
    
    set last [$root lastChild]
    
    append txt \n "// roots have lowest rank\n"  \{ " " rank=source \n 
    foreach const [$root selectNodes descendant::constellation] {
	
	append txt $const \n
	if { $const != $last } {
	    append txt ph0$const \n
	}
    }
    append txt \} \n\n
    
    foreach const [$root selectNodes descendant::constellation] {
	append txt [dotCodeForConst $const]
	append txt \n\n
	if { $const != $last } {
	    append txt "// dummy cluster to produce some horizontal space\n"
	    append txt "subgraph phantom$const " \{\n
	    append txt "ph0$const "
	    append txt {[label="", color=white] ;}  \n
	    for {set i 1} {$i<5} {incr i} {
		append txt "ph$i$const "
		append txt {[label="", color=white] ;}  \n
		append txt "ph[expr {$i-1}]$const -- ph$i$const " {[style=invis]} \n
	    }
	    append txt \}\n
	}
    }
    
    append txt \}

    return $txt
}


proc dotCodeForConst { const } {
    
    set rootname [$const @name]
    set rootNode [$const firstChild]
    
    # Now build the dot file:
    set txt "// constellation \"$rootname\" \n"
    append txt "subgraph \"cluster-$rootname\" "
    append txt \{ \n
    append txt color=white\n
    append txt ordering=out\n
    # Put the leaves at top:
    append txt "\{ rank=sink // these are the leaves:\n"
    foreach leaf [$const getElementsByTagName leaf] {
	append txt "$leaf // [$leaf @name] \n"
    }
    append txt \}\n
    append txt $const
    append txt " \[shape=none, label=\"$rootname\", color=white\]"
    append txt \n
    
    append txt "// nodes:\n"
    # All the nodes of the tree:
    foreach node [$const selectNodes descendant::*] {
	set name [$node @name]
	if { [$node nodeName] == "dot" } {
	    set characteristics \
	      "\[label=\"$name\", shape=circle, style=filled, color=grey\]"
	} elseif { [$node nodeName] == "leaf" } {
	    set characteristics \
	      "\[label=\"$name\", color=white\]"
	} else {
	    continue
	}
	append txt "$node $characteristics \n"
    }
    
    append txt {// edges:} \n
    # All the edges of the tree:
    foreach d [$const selectNodes descendant-or-self::*\[child::*\]] {
	foreach child [$d childNodes] {
	    # Draw the edge (using dom-node identifiers):
	    append txt "$d -- $child "
	    # If the child node is a stopdot, put its reference on the
	    # head of the edge:
	    if { ![$child hasChildNodes] && [$child hasAttribute ref] } {
		append txt "\[headlabel=\"[$child @ref]\"\]"
	    }
	    # Make a comment using the id-tags, to make the code readable:
	    append txt "// [$d @name] -- [$child @name]\n"
	}
    }
    
    append txt \}
    
    return $txt
}



# proc seeAll { pat {dir ""} } {
#     if { ![string length $dir] } {
# 	set dir [pwd]
#     }
#     
#     set L [glob -dir $dir -type f *$pat*.xml]
#     foreach file $L {
# 	set dotFile [opetope2dot $file]
# 	::dot::sendToGraphviz $dotFile
#     }
#     
#     set F [glob -nocomplain -dir $dir -type d *$pat*.faces]
#     foreach f $F {
# 	seeAll $pat $f
#     }
# }
# 
# proc seeOpetope { {xmlFile ""} } {
#     if { ![string length $xmlFile] } {
# 	set xmlFile [win::Current]
#     }
#     
#     set fd [open $xmlFile r]
#     set xml [read $fd]
#     close $fd
#     set doc [dom parse $xml]
#     set root [$doc documentElement]
#     ::dom::domNode::checkOpetope $root
# 
#     set dotFile [opetope2dot $xmlFile]
#     ::dot::sendToGraphviz $dotFile
# }



# Copyright (c) 2004-2006, Joachim Kock.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#    * Redistributions of source code must retain the above copyright 
#      notice, this list of conditions and the following disclaimer.
# 	   
#    * Redistributions in binary form must reproduce the above copyright 
#      notice, this list of conditions and the following disclaimer in the 
#      documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
