# -*-Tcl-*- (nowrap)
######################################################################
# file:        opetopes.tcl
# author:      Joachim Kock
# last update: 2006-10-17 14:51
######################################################################
# This file defines procedures for manipulating XML encoded opetopes.
# In particular, for computing target and source faces, and for computing
# composition.
# 
# The notion of opetope (invented by J.Baez and J.Dolan) and the algorithms
# implemented by the procedures refer to the paper 'Polynomial functors and
# opetopes' by J.Kock, A.Joyal, M.Batanin, and J.-F.Mascari.
# 
# 
# Copyright (c) 2004-2006, Joachim Kock.  All rights reserved.  
# Distributed under a BSD-style license (see bottom of file).
# 
# 
# Contents
# --------
# REMARKS
# SECTION 0  --  Preliminaries
# SECTION 1  --  Procedures for querying and checking
# SECTION 2  --  Procedures for sphere operations
# SECTION 3  --  Sources and target
# SECTION 4  --  Weights, normal form, and opetope comparison
# SECTION 5  --  Gluing
# SECTION 6  --  Reading and writing files
# SECTION 7  --  File-level procs
# SECTION 8  --  Some auxiliary procs


# --------------------------------------------------------------------
# REMARKS
# --------------------------------------------------------------------
# tDOM smartness: All procs defined in the ::dom::domNode namespace can be
# called also as methods for the domNode they otherwise would take as first
# argument.  In other words, a proc defined to have syntax
# 
#    ::dom::domNode::thisProc $thisNode args
#    
# can equally well be invoked as a method, through
# 
#    $thisNode thisProc args
# 
# just like standard domNode methods.


# --------------------------------------------------------------------
# SECTION 0  --  Preliminaries
# --------------------------------------------------------------------

if { [catch {package require tdom}] } {
    if { [info exists argv] } {
	# This file was sourced from a shell script.  Then we use puts instead of error:
	puts "ERROR: Could not load tDOM.  This package is essential, and nothing will work\n \
	  without it.  Please install tDOM.  See the Manual for instructions."
	exit 1
    } else {
	error "Could not load tDOM.  This package is essential, and nothing will work without it.\
	  Please install tDOM.  See the Manual for instructions."
    }
}



# In a debugging situation, increase the value to 1 or 10 to get
# some more info about in and out of key procs.
set DEBUGLEVEL 0
# The following should not be changed:
set DEBUGINDENT ""

# package require Tclx
# We only ever use two simple Tclx commands, which are easy to emulate:
proc lcontain {list element} {
    return [expr {([lsearch -exact $list $element] == -1) ? 0 : 1}]
}

proc lvarpop {var {indexExpr "0"} {string ""}} {
    upvar 1 $var L
    if {![info exists L]} {
	error "can't read \"${var}\": no such variable"
    }
    if {[regsub -- {^len} $indexExpr "" lenArg]} {
	set indexExpr [eval [list expr [llength $L]] $lenArg]
    } elseif {[regsub -- {^end} $indexExpr "" endArg]} {
	set indexExpr "end$endArg"
    }
    set result [lindex $L $indexExpr]
    set L [eval [list lreplace $L $indexExpr $indexExpr] $string]
    return $result
}




# The file opetope.dtd, needed pr parsing the XML files, should reside in
# ../xml/opetope.dtd, compared to the location of this file.  The problem
# is that this file doesn't know its own location.  We can try to figure it
# out: If the variable $homeOfThisScript is set (this is the case if the
# file is sourced from a shell script), then we rely on that one:
if { ![info exists homeOfThisScript] } {
    if { [namespace exists ::alpha] } {
	# If we are running this script from Alpha, then perhaps the file is
	# open in a window, in which case we can figure out where it came from.
	set path [lsearch -glob -inline [winNames -f] *opetopes.tcl]
	if { [string length $path] } {
	    set nameOfThisScript [file normalize $path]
	    set homeOfThisScript [file dirname $nameOfThisScript]
	}
	unset path
    } else {
	# If we still haven't found out, it might be that we are running in tclsh.
	# If this file was sourced through a [source] command, we can figure out
	# its path by looking in the history list!
	set zeroCmd [history event 0]
	puts $zeroCmd
	set i [lsearch -exact $zeroCmd "source"]
	if { $i != -1 } {
	    set filename [lindex $zeroCmd [expr {$i+1}]]
	    if { [file readable $filename] } {
		set nameOfThisScript [file normalize $filename] 
		puts $nameOfThisScript
		set homeOfThisScript [file dirname $nameOfThisScript] 
	    }
	}
    }
}


if { [info exists homeOfThisScript] } {
    set baseURI [file normalize [file join $homeOfThisScript .. xml]]
} else {
    # Let's see how far we get without a baseURI:
    set baseURI ""
    puts "WARNING: Could not figure out the location of this file."
    puts "Please help the script by writing in the beginning of it:"
    puts "    set homeOfThisScript /full/path/to/the/tcl-directory"
}





# --------------------------------------------------------------------
# SECTION 1  --  Procedures for querying and checking
# --------------------------------------------------------------------
# These procs do not modify the involved data.

proc ::dom::domNode::checkConstellation { upper } {
    set lower [$upper previousSibling]
    # First of all we check that the names are unique among dots:
    set dotNames [list]
    foreach D [$upper selectNodes descendant::dot] {
	lappend dotNames [$D @name]
    }
    if { $lower ne "" } {
	foreach D [$lower selectNodes descendant::dot] {
	    lappend dotNames [$D @name]
	}
    }
    if { [llength $dotNames] != [llength [lsort -unique $dotNames]] } {
	error "Dot names not unique"
    }
    # Next do the same for leaves:
    set leafNames [list]
    foreach D [$upper selectNodes descendant::leaf] {
	lappend leafNames [$D @name]
    }
    if { $lower ne "" } {
	foreach D [$lower selectNodes descendant::leaf] {
	    lappend leafNames [$D @name]
	}
    }
    if { [llength $leafNames] != [llength [lsort -unique $leafNames]] } {
	error "Leaf names not unique"
    }
    # Next we check that there are no stopdots if $upper is in fact lowest constellation:
    if { $lower eq "" } {
	# RULE: there may be no stopdots in the lowest constellation.
	foreach dotNode [$upper getElementsByTagName "dot"] {
	    if { [$dotNode isStopNode] } {
		error "Stopdots not allowed in lowest constellation."  
	    }
	}
	# There are no stopdots that might cause problems, and since there is no 
	# lower constellations, there is no zoom relation to check:
	return OK
    }
    
    set dotIDs [list]
    # First we check the bijection between dots in $lower and leaves in $upper:
    foreach N [$lower getElementsByTagName "dot"] {
	lappend dotIDs [$N @name]
    }
    set leafRefs [list]
    foreach leaf [$upper getElementsByTagName "leaf"] {
	lappend leafRefs [$leaf @name]
    }
    set dotIDs [lsort $dotIDs]
    set leafRefs [lsort $leafRefs]
    if { $dotIDs != $leafRefs } {
	error "No bijection between dots in \"[$lower @name]\" and leaves \"[$upper @name]\""
    }
    
    # Now check each of the dots in $upper:
    foreach dotNode [$upper getElementsByTagName "dot"] {
	if { [$dotNode isStopNode] } {
	    # For stopdots we must check that the referenced node exists:
	    set referencedNode [$lower find name [$dotNode @ref]]
	    if { ![string length $referencedNode] } {
		set referencedNode [$upper find name [$dotNode @ref]]
		if { ![string length $referencedNode] } {
		    error "Stopdot \"[$dotNode @name]\" doesn't refer to a valid dot"
		} elseif { ![isStopNode $referencedNode] } {
		    error "Stopdot \"[$dotNode @name]\" refers to non stopdot \"[$referencedNode @name]\""
		}
	    }
	} else {
	    # For non stopdots we must check that it defines a subtree in the lower
	    # constellation
	    set L [leavesOfNode $upper [$dotNode @name]]
	    if { [llength $L] } {
		# Otherwise there are no descendant leaves, and then the validity depends
		# of the descendant stopdots, which we checked above.
		set R [isSubtree $lower $L] ;# this will throw error if not a subtree
	    }
	}
    }
    return OK
}


proc ::dom::domNode::checkOpetope { op } {
    # Check that dot names are unique:
    set dotNames [list]
    foreach D [$op selectNodes descendant::dot] {
	lappend dotNames [$D @name]
    }
    while { [llength $dotNames] } {
	set thisName [lvarpop dotNames]
	if { [lcontain $dotNames $thisName] } {
	    error "Dot names not unique: name \"$thisName\" used more than once."
	}
    }
    
    if { [llength $dotNames] != [llength [lsort -unique $dotNames]] } {
	error "Dot names not unique"
    }
    # Check that leaf names are unique:
    set leafNames [list]
    foreach D [$op selectNodes descendant::leaf] {
	lappend leafNames [$D @name]
    }
    while { [llength $leafNames] } {
	set thisName [lvarpop leafNames]
	if { [lcontain $leafNames $thisName] } {
	    error "Leaf names not unique: name \"$thisName\" used more than once."
	}
    }
    # Now check zoom relations and such (cf. the proc [checkConstellation]):
    foreach const [$op childNodes] {
	if { [catch {checkConstellation $const} err] } {
	    error "Non-valid opetope: problem in constellation \"[$const @name]\": $err"
	}
    }
    return OK
}


# The general rule for managing stopdot references is:
# Whenever we leave out a stopdot, contract it before leaving it
# out.  Then the only proc to take care of stopdot references is
# [contracNode].  In other words, we actually never leave out a stopdot
# --- we always contract it instead.




# ...including the node itself if it is a leaf.  In fact all descendant leaves!
proc ::dom::domNode::leavesOfNode { const dotID } {
    set n [$const getNodeWithName $dotID]   
#     # The following might be done more elegantly
    set dotIDs [list]
    foreach desc [$n selectNodes descendant-or-self::leaf] {
	lappend dotIDs [$desc @name]
    }
    # by doing something like this
#     set dotIDs [$n selectNodes descendant-or-self::leaf/@name]
    return $dotIDs
}




# If it is a contracting subtree, return the name of the root.
# Or throw an error if the list of nodes do not form a subtree.
# We characterise a subtree as a nonempty collection of nodes (dots or
# leaves) such that only one of the nodes does not have its parent in the
# collection.
# 
# Note that we do not allow those subtrees consting of a single edge,
# although these are actually valid subtrees in the technical sense.  In
# fact there is no way to specify such subtrees, since we haven't even got
# names for the edges, so no callers of this proc will ever come up with
# such dotless subtrees...  
# 
# Well in fact we do allow subtrees consisting of a single leaf!
# 
proc ::dom::domNode::isSubtree { const dotIDs } {
    if { $::DEBUGLEVEL > 9 } {
	puts "${::DEBUGINDENT}ENTER ::dom::domNode::isSubtree $const $dotIDs.  ($const has name \"[$const @name]\")"
	append ::DEBUGINDENT " "
    }
    if { ![llength $dotIDs] } {
	error "There is no such thing as the empty tree"
    }
    set L [list] ;# will be the list of all nodes in the collection
    foreach thisID $dotIDs {
	lappend L [$const getNodeWithName $thisID]
    }
    set R [list] ;# will be the sublist of those nodes in $L whose parent is not in $L
    foreach node $L { 
	if { ![lcontain $L [$node parentNode]] } {
	    lappend R $node
	} 
    }
    if { [llength $R] == 1 } {
	# That's the characterisation of subtree, and the only entry in $R
	# is the root node:
	set R [lindex $R 0]
	if { $::DEBUGLEVEL > 9 } {
	    set ::DEBUGINDENT [string range $::DEBUGINDENT 0 end-1]
	    puts "${::DEBUGINDENT}RES = $R"
	}
	return [$R @name]
    } elseif { [llength $R] == 0 } {
	# This can never happen:
	error "This can never happen"
    } else {
	error "Dots $dotIDs do not form a subtree"
    }
}


# Check if a given node is a stopnode.  If it is, return 1; if not, return 0.
# If it is a childless dot but has no ref attribute, throw an error:
proc ::dom::domNode::isStopNode { N } {
    if { $::DEBUGLEVEL > 9 } {
	puts "${::DEBUGINDENT}ENTER ::dom::domNode::isStopNode $N.  ($N has name \"[$N @name]\")"
	append ::DEBUGINDENT " "
    }
    set res 1
    if { [$N nodeName] ne "dot"} {
	# It is not even a dot!
	set res 0
    } elseif { [$N hasChildNodes] } {
	# This dot has children!  If it has a ref attribute, that's a syntax error!:
	if { [$N hasAttribute ref] } {
	    error "Dot \"[$N @name]\" is not a stopdot, but has a reference tag"	    
	}
	set res 0
    } elseif { ![$N hasAttribute ref] } {
	# It is in fact a stopnode, since it has no children
	# but on the other hand it doesn't have a ref attribute!
	# This is a syntax error!
	error "No ref attribute for stopnode $N (stopdot \"[$N @name]\")"
    }
    if { $::DEBUGLEVEL > 9 } {
	set ::DEBUGINDENT [string range $::DEBUGINDENT 0 end-1]
	puts "${::DEBUGINDENT}RES = $res"
    }
    return $res
}


# Find the node that $stopNode refers to.
proc ::dom::domNode::pointsTo { stopNode } {
    if { ![$stopNode isStopNode] } {
	error "\"[$stopNode @name]\" is not a stop dot"
    }
    set upper [$stopNode selectNodes ancestor::constellation]
    set lower [$upper previousSibling]
    set ref [$stopNode @ref]
    # First look for the node in $lower:
    set referencedNode [$lower find name $ref]
    if { ![string length $referencedNode] } {
	# We didn't find anything below, so try in current constellation:
	set referencedNode [$upper find name $ref]
	if { ![string length $referencedNode] } {
	    error "Stopdot \"[$stopNode @name]\" doesn't refer to a valid dot"
	} elseif { ![isStopNode $referencedNode] } {
	    error "Stopdot \"[$stopNode @name]\" refers to non stopdot \"[$referencedNode @name]\""
	}
    }
    return $referencedNode
}



# Given a stopdot (by name), find out which node in the constellation below
# it refers to, after following all stopdots in the current constellation.
# The return value is a dotID!
proc ::dom::domNode::lastStepInReferChain { const stopdotName } {
    if { $::DEBUGLEVEL > 9 } {
	puts "${::DEBUGINDENT}ENTER ::dom::domNode::lastStepInReferChain $const $stopdotName.  ($const has name [$const @name])"
	append ::DEBUGINDENT " "
    }
    set lower [$const previousSibling]
    if { ![string length $lower] } {
	# This should never happen: stopdots can not occur in bottom constellation.
	error "Stopdots not allowed in bottom constellation"
    }
    # Find the domNode name of the given stopdot:
    set node [$const getNodeWithName $stopdotName]
    if { ![isStopNode $node] } {
	error "(in lastStepInReferChain): \"$stopdotName\" is not a stopdot"
    }
    # See what it refers to:
    set reference [$node @ref]
    # Look for such a node in the tree below:
    set foundNode [$lower find name $reference]
    # If we found something, then this is the final answer.
    # Otherwise, we'll have to look in the current tree instead:
    while { ![string length $foundNode] } {
	# Look for the node in the current tree:
	set foundNode [$const find name $reference]
	# Stopdots may only refer to stopdots!
	if { ![isStopNode $foundNode] } {
	    error "(in lastStepInReferChain): stopdot \"$stopdotName\" refers to a non-stopdot \"[$foundNode @name]\""
	}
	# See what it refers to:
	set reference [$foundNode @ref]
	# And look for such a node in the tree below:
	set foundNode [$lower find name $reference]
    }
    # The final answer is the name of a node in the three below.
    if { $::DEBUGLEVEL > 9 } {
	set ::DEBUGINDENT [string range $::DEBUGINDENT 0 end-1]
	puts "${::DEBUGINDENT}RES = [$foundNode @name]"
    }
    return [$foundNode @name]   
}




# Find a stopnode in $const referring to the name $name:
proc ::dom::domNode::findStopnodeReferringToName { const name } {
    set res [$const selectNodes descendant::dot\[@ref='$name'\]]
    if { [llength $res] > 1 } {
	error "More than one stopdot in \"[$const @name]\" referring to \"$name\""
    } elseif { [llength $res] == 0 } {
        return ""
    } else {
	return [lindex $res 0]
    }
}





# # List of nodes referring to the name of the given dot.
# # These nodes are ANYWHERE in the opetope, not just in the
# # specified constellation.
# proc ::dom::domNode::listNodesReferringToThisName { const name } {
#     return [$const selectNodes //dot\[@ref='$name'\]]
# }
# This previous version of the proc returned a list of all stopnodes in
# $doc referring to $name.  In theory there might be two references: one to
# a dot of that name, and one to a leaf of that name.  However these two
# referring stopdots can then clearly not be in the same constellation!  It
# turns out that the callers of the proc know precisely in which
# constellation they want to find the stopdot, which is precisely the
# argument passed to this proc.  So the new version will search for the
# stopdot only in $const, not in the whole $doc.







# It is too common an error to refer to a node by its name,
# and then the name is wrong, so no node exists with this name.
# The standard method for getting the node by its dotID is
# something like
# 
#     set N [$const find name $dotID]
#     
# and then continue using this domNode, for example in
# 
#     $N firstChild
#     
# However, this will throw an error if $dotID was not found,
# because in that case the first statement will assign the
# empty string to $N, and obviously we can't use the empty
# string as a command.  The problem is that the error message
# 
#    Error: invalid command name ""
# 
# is rather useless.  To remedy this situation, we always use
# the wrapper [::dom::domNode::getNodeWithName] which is
# synonymous with the first command above, except that it
# raises an error is nothing is found...
proc ::dom::domNode::getNodeWithName { const name } {
    set N [$const find name $name]
    if { $N eq "" } {
	error "No node named \"$name\" in constellation \"[$const @name]\""
    }
    return $N    
}



# --------------------------------------------------------------------
# SECTION 2  --  Procedures for sphere operations
# --------------------------------------------------------------------
# 
# There are four main sphere operations that change the opetope:
# 
#    Sphere operation                   operation on composition tree
#    ----------------------------------------------------------------
#    contract a sphere to a dot         prune a dot
#    erase a non-outer sphere           contract a dot back to parent
#    draw sphere around sphere or dot   insert dot under dot or leaf
#    restrict to a sphere               select saturated subtree
# 
# Each of these operations imply repercussions down in the lower
# constellations.
# 
# We have the following four corresponding procs
# 
#    ::dom::domNode::prune
#    ::dom::domNode::contractNode
#    ::dom::domNode::newDotUnder
#    ::dom::domNode::selectSaturatedSubtree
#    
# It is important that each of these procs check that the zoom relation is
# not spoiled, so each of these procs ends by invoking [checkConstellation]
# before returning.  This is not per se a proof of correctness, but it is an
# important guard agains silly mistakes.
# 
# 
# There are also variant procs: 
# 
#    ::dom::domNode::contractTree
#         given a list of dots, contract them one by one (so it's a 
# 	wrapper for [::dom::domNode::contractNode])
#    ::dom::domNode::selectSubtree
#         first prune the descendant dots not contained in the subtree, and
# 	then select the resulting saturated subtree
#    
# Since these are just wrappers for the first four, we don't need to have
# strict type checking in these two.


# When pruning a dot (with name $dotID) we are replacing the
# saturated-subtree-with-root-$dotID by a leaf.  Ths leaf should get name $dotID.
proc ::dom::domNode::prune { const dotID } {
    if { $::DEBUGLEVEL } {
	puts "${::DEBUGINDENT}ENTER ::dom::domNode::prune $const $dotID.  ($const has name \"[$const @name]\")"
	append ::DEBUGINDENT " "
    }
    set N [$const getNodeWithName $dotID]

    # Before actually pruning, it is better to contract all descendant stopdots
    # just to resolve all references properly:
    set stopnodeList [$N selectNodes descendant::dot\[@ref\]]
    while { [llength $stopnodeList] } {
	foreach stopdot $stopnodeList {
	    contractNode $const $stopdot
	}
	# Contracting a stopdot back to dot K may turn K into a stopdot,
	# so we need to renew the list:
	set stopnodeList [$N selectNodes descendant::dot\[@ref\]]
    }

    # The pruning will consist in replacing the parent node by a leaf:
    set p [$N parentNode]
    set doc [$const ownerDocument]
    set newLeaf [$doc createElement leaf]
    # The resulting leaf should get the name of the dot we pruned.
    # Correspondingly, in the lower tree, we will contract a circle,
    # and the resulting dot should have this same name:
    set leafName $dotID
    # Now we should just make sure there are no references in $const to any
    # of the dots below we will contract, including the root dot of that
    # contraction.  References to any of the non-root dots in that contraction
    # would correspond to empty spheres inside the sphere we will contract, and 
    # hence the stopdot up in $const would be in the tree we are pruning, so these
    # go away and do not present a problem.  Hence the only possibility is that 
    # there is some stopdot in $const referring to the root dot of the contracting
    # tree in $lower.  So look in $const if there is a reference to the root of
    # the contracting subtree:
    set lower [$const previousSibling]
    set L [leavesOfNode $const $dotID]
    if { [llength $L] && [string length $lower] } {
	set oldRootName [isSubtree $lower $L]
	# perhaps we ought to check here that this is indeed a node...
	if { ![string length $oldRootName] } {
	    error "somehow we are going to try contract a tree without a root..."
	}
	foreach r [findStopnodeReferringToName $const $oldRootName] {
	    $r setAttribute ref $dotID
	}
    }

    # Finally, in the lower constellations we have to contract the
    # corresponding subtree.  However, we first treat the case where there
    # is no lower constellation:
    if { $lower eq "" } {
	$newLeaf setAttribute name $leafName
	$p insertBefore $newLeaf $N
	$N delete
	if { $::DEBUGLEVEL } {
	    set ::DEBUGINDENT [string range $::DEBUGINDENT 0 end-1]
	    puts "${::DEBUGINDENT}LEAVE ::dom::domNode::prune $const $dotID.  ($const has name \"[$const @name]\")"
	}
	return
    }

    if { [llength $L] } {
	contractTree $lower $L $leafName
    } else {	
	# In this case is the same as saying that the node is itself a stopdot!
	# Because previously we contracted descendant stopdots!
	if { ![$N isStopNode] } {
	    error "(in ::dom::domNode::prune): expected dot \"[$N @name]\" to be \
	      a stopdot, but it isn't"
	}
	
	# Although the node was removed, we can still access its data:
	set ref [lastStepInReferChain $const [$N @name]]
	set belowWhatNode [$lower find name $ref]
	if { ![string length $belowWhatNode] } {
	    error "(in ::dom::domNode::prune): could not find ref $ref"
	}
	
	newDotUnder $lower $belowWhatNode $dotID

    }

    $newLeaf setAttribute name $leafName
    $p insertBefore $newLeaf $N
    $N delete
    # Before returning, check that the resulting opetope is well formed:
    if { [catch {$const checkConstellation} err] } {
	error "pruning \"[$const @name]\" at $dotID spoiled zoom relation ($err)"
    }
    if { $::DEBUGLEVEL } {
	set ::DEBUGINDENT [string range $::DEBUGINDENT 0 end-1]
	puts "${::DEBUGINDENT}LEAVE ::dom::domNode::prune $const $dotID.  ($const has name \"[$const @name]\")"
    }
}


proc ::dom::domNode::newDotUnder { const belowThisNode dotID } {
    set doc [$const ownerDocument]
    set new [$doc createElement dot]
    $new setAttribute name $dotID
    set p [$belowThisNode parentNode]
    $p replaceChild $new $belowThisNode
    # MYSTERY: WHY DOESN'T IT WORK TO DO [insert $belowThisNode] as child of $new?
    $new appendXML [$belowThisNode asXML]
    # Since we have modified this constellation, we ought to check that it is
    # still valid.  However, setting a new dot never spoils zoom relation.
}

# # Contract a dot back to its parent dot:
# proc ::dom::domNode::contractDot { const dotName } {
#     set dotNode [$const getNodeWithName $dotName]
#     contractNode $const $dotNode
# }



# Currently we don't allow to contract a leaf.  It doens't really make sense.
proc ::dom::domNode::contractNode { const N } {
    if { $::DEBUGLEVEL } {
	puts "${::DEBUGINDENT}ENTER ::dom::domNode::contractNode $const $N.  ($const has name \"[$const @name]\"; $N has name \"[$N @name]\")"
	append ::DEBUGINDENT " "
    }
    set p [$N parentNode]
    if { [$N nodeName] ne "dot" } {
	error "can't contract this node $N (it is a [$N nodeName] \"[$N @name]\")"
    }
    if { [$p nodeName] ne "dot" } {
	error "can't contract this node $N (its parent is a [$p nodeName] \"[$p @name]\")"
    }
    # Special case: the dot is a stopdot.  There are two issues:
    # 1) it might be that some other dot in this same cosntellation is
    # referring to it, i.e. sits above it in the refer chain.  Note that we
    # don't care about if a dot in the higher constellation refers to it,
    # because if we are contracting a node $N here in $const, then it is
    # because we pruned something in the higher constellation, and any
    # stopdot up there that could have referred to $N$ is already gone at
    # this point.  Hence we should only look for references in the current
    # constellation.
    # 2) If $N is the only child of $p, then $p becomes a new stopdot, and
    # it should inherit the ref from $N.
    if { [$N isStopNode] } {
	$p removeChild $N
	set SNR [findStopnodeReferringToName $const [$N @name]]
	if { ![$p hasChildNodes] } {
	    # $p is now a stopdot.  Observe that we couldn't just have
	    # asked whether $p is a stopdot, because we have not yet given
	    # it a ref attribute!.  In fact we are precisely going to give
	    # this attribute now:
	    $p setAttribute ref [$N @ref]
	    # If there is a stopdot referring to $N, then it should just
	    # refer to $p instead:
	    if { [string length $SNR] } {
		$SNR setAttribute ref [$p @name]
	    }
	} else {
	    # If $p does not become a stopdot, then $SNR should just skip
	    # $N in the refer chain.
	    if { [string length $SNR] } {
		$SNR setAttribute ref [$N @ref]
	    }
	}
    } else {
	# This is the standard case: take each child out of $N and stick
	# it directly into $p:
	foreach child [$N childNodes] {
	    $N removeChild $child
	    $p insertBefore $child $N
	}
    }
    
    $N delete
    # Before returning, check that the resulting opetope is well formed:
    if { [catch {$const checkConstellation} err] } {
	error "contracting node \"[$N @name]\" in constellation \"[$const @name]\" spoiled zoom relation ($err)"
    }
    if { $::DEBUGLEVEL } {
	set ::DEBUGINDENT [string range $::DEBUGINDENT 0 end-1]
	puts "${::DEBUGINDENT}LEAVE ::dom::domNode::contractNode $const $N." ;#  ($const has name \"[$const @name]\"; $N has name \"[$N @name]\")"
    }
    return $p
}

# Accept also the situation where there is only one node in the
# tree.  In that case the contraction is the identity.
# On the other hand, we do not allow dotless subtrees.  (Contracting
# a dotless tree is the same thing as setting a new dot, and we
# have another proc to do this.)
proc ::dom::domNode::contractTree { const L {newName ""} } {
    set R [isSubtree $const $L]
    # $R is the ID of the root dot
    foreach dotID $L {
	if { $dotID ne $R } {
	    set dotNode [$const getNodeWithName $dotID]
	    contractNode $const $dotNode
	    # this will take care of what happens in lower constellations
	}
    }
    if { [string length $newName] } {
	set survivingDotNode [$const getNodeWithName $R]
	
	# We don't check whether anybody referred to any of these nodes
	# (it could only be to the root anyway).  This task is up to the
	# caller (prune) from the upper tree...

	$survivingDotNode setAttribute name $newName
    }
}

# Select saturated subtree:
# Throw away everything above the node named $rootnameOfSubtree.
# (I.e. nearer the root.)  This is the opposite of pruning.
# 
# Note that we accept also the saturated subtree consisting of a single leaf.
# 
proc ::dom::domNode::selectSaturatedSubtree { const rootnameOfSubtree } {
    if { $::DEBUGLEVEL } {
	puts "${::DEBUGINDENT}ENTER ::dom::domNode::selectSaturatedSubtree $const $rootnameOfSubtree. ($const has name \"[$const @name]\")"
	append ::DEBUGINDENT " "
    }
    set newRootNode [$const getNodeWithName $rootnameOfSubtree]
    set lower [$const previousSibling]
    # We are going simply to replace the original root node by $newRootNode.
    # But first we have to make the necessary changes in the lower constellations,
    # if there are any:
    if { [string length $lower] } {
	# First we have to take care of any stopdots outside the subtree, so that
	# we don't loose their references.  We simply contract them away.  The proc contractNode is designed
	# to handle stopdot references, so in this way we take care of that.
	set allStopNodes [$const selectNodes descendant-or-self::dot\[@ref\]]
	set descStopNodes [$newRootNode selectNodes descendant-or-self::dot\[@ref\]]
	set otherStopNodes [list]
	foreach stopNode $allStopNodes {
	    if { ![lcontain $descStopNodes $stopNode] } {
		lappend otherStopNodes $stopNode
	    }
	}

	foreach stopNode $otherStopNodes {
	    set P [contractNode $const $stopNode]
	    # If the parent node becomes a stopnode, contract it too, and so on:
	    while { [$P isStopNode] } {
		set P [contractNode $const $P]
	    }
	}
	
	set L [leavesOfNode $const $rootnameOfSubtree]
	if { ![llength $L] } {
	    # The dots are all stopdots!
	    # All stopdots are descendants of $rootnameOfSubtree...
	    # But there might be many stopdots, referring to different
	    # things down in $lower!?!  MYSTERY!  SOMETHING IS WRONG HERE!
	    # 
	    # HERE IS THE ARGUMENT: $rootnameOfSubtree has no descendant
	    # leaves.  Hence the corresponding sphere in $lower cuts only
	    # an edge.  All stopdots under $rootnameOfSubtree correpond
	    # to spheres on that edge, arranged in linear order.  Hence
	    # it is guaranteed that all the stopdots refer to each other
	    # in linear order, and that there is only one final reference
	    # to $lower.  Hence, to find this reference, we can take any
	    # stopdot under $rootnameOfSubtree (possibly $rootnameOfSubtree
	    # itlsef), and follow its ref chain:
	    set someStopNode [lindex $descStopNodes 0]
	    set someStopDot [$someStopNode @name]
	    # 
	    # Find out which node they refer to.  We don't care
	    # if they refer to some other stopdot in the same tree,
	    # because we are selecting something else, and those
	    # other stopdots will be thrown away.  We only care for
	    # the final reference down in the lower tree:
	    set ref [lastStepInReferChain $const $someStopDot]
	    # Down in $lower, we are going to select a tree without
	    # any dots.  So all we need to know is what the name of
	    # the leaf of that little subtree will be.  To get this
	    # leaf, we first prune the dot referenced, and then
	    # select the subtree whose only node is the resulting
	    # leaf:
	    prune $lower $ref

	    set L [list $ref]
	}
	# Finally, select the tree below:
	selectSubtree $lower $L
    }
    
    # Finish the operation in the upper tree, deleting
    # everything else but the selected subtree:
    set originalRootNode [$const firstChild]
    $const replaceChild $newRootNode $originalRootNode
    # Before returning, check that the resulting opetope is well formed:
    if { [catch {$const checkConstellation} err] } {
	error "zoom relation spoiled by ::dom::domNode::selectSaturatedSubtree $const $rootnameOfSubtree. ($const has name \"[$const @name]\")"
    }
    if { $::DEBUGLEVEL } {
	set ::DEBUGINDENT [string range $::DEBUGINDENT 0 end-1]
	puts "${::DEBUGINDENT}LEAVE ::dom::domNode::selectSaturatedSubtree $const $rootnameOfSubtree. ($const has name \"[$const @name]\")"
    }
}




# The list L may not be empty.  (If it is, an error wil be thrown by [isSubtree].)
# Of course it is understood that if a dot is included then all its leaves are also included.
proc ::dom::domNode::selectSubtree { const L } {
    if { $::DEBUGLEVEL } {
	puts "${::DEBUGINDENT}ENTER ::dom::domNode::selectSubtree $const $L.  ($const has name \"[$const @name]\")"
	append ::DEBUGINDENT " "
    }
    # Get the root node of the collection:
    set Rname [isSubtree $const $L]
    set newRootNode [$const find name $Rname]
    # Prune the descendant dots not contained in the subtree:
    set descDots [$newRootNode selectNodes descendant::dot]
    foreach dot $descDots {
	if { [catch {set dotName [$dot @name]}] } {
	    # This happens because we are looking at a dot that was
	    # already pruned away!
	    continue
	}
	if { ![lcontain $L $dotName] } {
	    prune $const $dotName
	}
    }
    # Finally contract the tree which is now saturated:
    selectSaturatedSubtree $const $Rname
    if { $::DEBUGLEVEL } {
	set ::DEBUGINDENT [string range $::DEBUGINDENT 0 end-1]
	puts "${::DEBUGINDENT}LEAVE ::dom::domNode::selectSubtree $const $L.  ($const has name \"[$const @name]\")"
    }
}


# --------------------------------------------------------------------
# SECTION 3  --  Sources and target
# --------------------------------------------------------------------
# Since the sphere operation procs modify the opetope they operate on,
# and since in practice we would like to keep the original opetope in
# memory, even if we compute a source of it, when computing facets, we
# should operate on a copy.

proc ::dom::domNode::cloneOpetopeNode { opetopeNode } {
#      # Serialisation way of cloning:
#     append xml {<?xml version="1.0" encoding="UTF-8"?>} \n
#     append xml {<!DOCTYPE opetope SYSTEM "opetope.dtd">} \n
#     append xml [$opetopeNode asXML]
#     set newDoc [dom parse $xml]
#     set newOpetopeNode [$newDoc documentElement]
#     # Now we've got a brand new copy of the opetope.
    
    # Non-serialisation way of cloning:
    set newOpetopeNode [$opetopeNode cloneNode -deep]
    # Now we've got a brand new copy of the opetope.  However, currently
    # it is only present in the fragment list of the old owner document.
    # We would rather have it in a document on its own -- this is safer
    # in case we later finish with the original opetope and delete its
    # owner...  Hence:
    set newDoc [dom createDocumentNode]
    $newDoc systemId opetope.dtd
    $newDoc appendChild $newOpetopeNode
    # From now on, the new opetope node belongs to this new document,
    # which can be accessed through [$newOpetopeNode ownerDocument].
    return $newOpetopeNode    
}


# Given an opetope node and a name of a source of it, produce that source
# opetope, (a new node in a new owner document).  A handle to the new
# opetope is returned.
proc ::dom::domNode::Source { opetopeNode sourceName } {
    # We assume the opetope is already valid, so no check at this point.
    set originalOpetopeName [$opetopeNode @name]
    set newOpetopeNode [$opetopeNode cloneOpetopeNode]  
    set upper [$newOpetopeNode lastChild]
    # Perform the source operation:
    set N [$upper find name $sourceName]
    if { $N eq "" || [$N nodeName] ne "dot"} {
	error "\"[$sourceName @name]\" is not a source of opetope \"$originalOpetopeName\". \
	  The sources are given by the dots in the composition tree. \
	  Use the proc \"listSources\" to get a list of all sources."
    } 
    set newOpetopeName "${originalOpetopeName}-[$N @name]"
    $newOpetopeNode setAttribute name $newOpetopeName
    $upper selectSubtree [list $sourceName]
    $upper delete
    # Finally check:
    if { [catch {$newOpetopeNode checkOpetope} err] } {
	# If an error occurred, we might as well clean up.
	# In any case we wouldn't even have a handle to this
	# messy new thing we created...
	set newDoc [$newOpetopeNode ownerDocument]
	$newDoc delete
	error "Failed to compute source: $err"
    }
    return $newOpetopeNode
}

# Given an opetope node, produce the target opetope, (a new node in a 
# new owner document).  A handle to the new opetope is returned.
proc ::dom::domNode::Target { opetopeNode } {
    # We assume the opetope is already valid, so no check at this point.
    set originalOpetopeName [$opetopeNode @name]
    set newOpetopeNode [$opetopeNode cloneOpetopeNode]  
    set upper [$newOpetopeNode lastChild]
    # Perform the target operation:
    $upper delete
    set newOpetopeName "${originalOpetopeName}-%"
    $newOpetopeNode setAttribute name $newOpetopeName
    # Finally check:
    if { [catch {$newOpetopeNode checkOpetope} err] } {
	# If an error occurred, we might as well clean up.
	# In any case we wouldn't even have a handle to this
	# messy new thing we created...
	set newDoc [$newOpetopeNode ownerDocument]
	$newDoc delete
	error "Failed to compute target $err"
    }
    return $newOpetopeNode
}



# Return the names of the sources (i.e. dots in the last
# constellation (the composition tree)):
proc ::dom::domNode::listSourceNames { opetopeNode } {
    set upper [$opetopeNode lastChild]
    set dotList [$upper selectNodes descendant-or-self::dot]
    set dotNames [list]
    foreach dot $dotList {
	lappend dotNames [$dot @name]
    }
    return $dotNames
}


# --------------------------------------------------------------------
# SECTION 4  --  Weights, normal form, and opetope comparison
# --------------------------------------------------------------------

# In this section we assign weights to each node of an opetope, and get a
# sort of normal form for opetopes, so that they can be compared even if
# their name attributes differ.
# 
# Each node is associated a weight: A weight is defined to be a vector of
# natural numbers: The weight of a node in dimension n is a vector of
# dimension n+1.  (I.e. if the tree appears as the third constellation in a
# zoom complex (dim 2), then the weights are lists of length 3.)  The first
# entry in the vector counts how many dots lie below the node (including
# itself).  The remaining entries of the vector is the vector sum of the
# weights of the dots referred to by the leaves below the node.
# 
# It is quite unlikely that two nodes have the same weight, but it is
# not excluded that it can happen.
# 
# An opetope is on normal form when the children of each dot are sorted
# according to weight.  Now we can compare opetopes by comparing their
# normal form.  Since the normal form is not unique, it is not guaranteed
# that we discover if two opetopes are isomoprhic, but if the algorithm
# says they are then it is guaranteed that they really are (and a specific
# isomorphism is provided).


# Compute the weight of a given node, and write it as weight attribute
proc ::dom::domNode::assignWeight { domNode } {
    if { [$domNode nodeName] eq "constellation" } {
	assignWeight [$domNode lastChild]
	return
    } elseif { [$domNode nodeName] eq "opetope" } {
	foreach const [$domNode childNodes] {
	    $const assignWeight
	}
	return
    } 
    set upper [$domNode selectNodes ancestor::constellation]
    set lower [$upper previousSibling]
    set previousConstellationList [$upper selectNodes preceding-sibling::*]
    set dim [llength $previousConstellationList]
    if { [$domNode nodeName] eq "leaf" } {
	# The weight of a leaf is the weight of the dot it refers to
	# in the previous constellation, with a 0 prepended to get the
	# correct dimension.
	if { $lower ne "" } {
	    set ref [$domNode @name]
	    set correspondingDotNode [$lower getNodeWithName $ref]
	    set w [linsert [$correspondingDotNode getWeight] 0 0]
	    $domNode setAttribute weight $w
	    return $w
	} else {
	    # we are in lowest dimension!
	    set w [list 1]
	    $domNode setAttribute weight $w
	    return $w
	}
    } elseif { [$domNode nodeName] eq "dot" } {
	# Start with the weight {1 0 0 ... 0}
	set w [linsert [lrepeatAllowingZero $dim 0] 0 1]
	set childList [$domNode childNodes]
	# If there are no child nodes, then this first weight is correct:
	if { ![llength $childList] } {
	    $domNode setAttribute weight $w
	    return $w
	}
	# Otherwise, add the weight of the children:
	set listOfChildWeights [list $w]
	foreach childNode $childList {
	    lappend listOfChildWeights [$childNode getWeight]
	}
	set w [sumWeights $listOfChildWeights]
	$domNode setAttribute weight $w
	return $w
    }
}


# We should never use [$domNode @weight], because a typical opetope
# doesn't have weight data.  Instead, use always the following proc
# which assigns the weight if it doesn't exist, computing all lower
# weights recursively when necessary.
proc ::dom::domNode::getWeight { domNode } {
    if { [catch {$domNode @weight} weight] } {
	set weight [$domNode assignWeight]
    }
    return $weight
}




# Given a domNode, sort its children according to weight.  
# (The weights are computed and assigned if they do not already exist.)
proc ::dom::domNode::sortChildren { domNode } {
    # Build a list whose items are pairs consisting of 
    # a domNode token and a weight. Then sort this list
    # according to the second part of each item.
    set L [list]
    foreach child [$domNode childNodes] {
	lappend L [list $child [split [$child getWeight]]]
	$domNode removeChild $child
    }
    foreach pair [lsort -command comparePairsByWeight $L] {
	set child [lindex $pair 0]
	$domNode appendChild $child
	$child setAttribute weight [lindex $pair 1]	
    }
}


# Two auxiliary procs used in [lsort]
proc comparePairsByWeight { P0 P1 } {
    return [compareWeights [lindex $P0 1] [lindex $P1 1]]
}
# In the next compare proc, we use the opposite convention than usually,
# because we want the list to be sorted in decreasing order.
proc compareWeights { w0 w1 } {
    if { [llength $w0] == 0 } {
	return 0
    }
    set n0 [lvarpop w0]
    set n1 [lvarpop w1]
    if { $n0 < $n1 } {
	return 1
    } elseif { $n1 < $n0 } {
        return -1
    } else {
        return [compareWeights $w0 $w1]
    }   
}


# Normalise all descendants.
proc ::dom::domNode::normaliseByWeight {domNode } {
   foreach dotNode [$domNode getElementsByTagName "dot"] {
       $dotNode sortChildren
   } 
}

# We want to check whether two opetopes are isomorphic: just
# assign weights and normalise, then replace all refs by xpath
# tags, and finally zap all name attributes.  Then compare the
# resulting xml:
proc ::dom::domNode::areIsomorphic { opNode0 opNode1 {bijectionMap ""} } {
    # First we check if there is an obvious obstruction to having an isomorphism:
    set res [string equal [$opNode0 asAnonymousXML] [$opNode1 asAnonymousXML]]
    if { !$res } {
	return 0
    }
    # If there isn't, then we go on to actually constructing an isomorphism:
    set op0 [$opNode0 cloneOpetopeNode]
    $op0 normaliseByWeight 
    set op1 [$opNode1 cloneOpetopeNode]
    $op1 normaliseByWeight 
    # Check if a variable name for the bijection map was supplied by the caller:
    if { $bijectionMap ne "" } {
	upvar 1 $bijectionMap bijMap
    }
    # We build up the bijection map little by little:
    set map [list]
    set listOfPairs [list]
    foreach N0 [$op0 selectNodes descendant-or-self::*] {
	set xp [$N0 toXPath]
	set N1 [$op1 selectNodes $xp]
	if { [lcontain $listOfPairs [list [$N0 @name] [$N1 @name]]] } {
	    # The pair is already in the map, so don't put it again:
	    continue
	}
	lappend listOfPairs [list [$N0 @name] [$N1 @name]]
	# Now we know the pair is not an old one.  We still don't know if
	# it is compatible with the map we have so far: we ought to check
	# whether the first entry appears in the odd part and the second
	# entry appears in the even part.  What we actually do, by laziness,
	# is just to try: append the pair to the map...
	lappend map [$N0 @name] [$N1 @name]
	# ... and then immediately check if this was correct:
	if { ![isBijectionMap $map] } {
	    return 0
	}
    }
    # At this point we have a bijection map on the level of names.
    # We don't need to check that the bijection extends to the refs,
    # because we have already checked that the two opetopes are
    # anonymously isomorphic.
    # 
    # We finish by simplify the map
    set simplerMap [list]
    foreach {x0 x1} $map {
	if { $x0 ne $x1 } {
	    lappend simplerMap $x0 $x1
	}
    }    
    set bijMap $simplerMap
    # Clean up before returning:
    set doc0 [$op0 ownerDocument]
    $doc0 delete
    set doc1 [$op1 ownerDocument]
    $doc1 delete
    return $res
}


# Return the XML of a opetope document, but normalised by weight, with all
# names stripped, and with references replaced by xpath style refs :
proc ::dom::domNode::asAnonymousXML { opNode } {
    set OP [$opNode cloneOpetopeNode]
    $OP normaliseByWeight
    # substitute every ref attribute with the xpath specification of the
    # node it points to
    foreach stopNode [$OP selectNodes descendant::dot\[@ref\]] {
	set referredNode [$stopNode pointsTo]
	set xp [$referredNode toXPath]
	$stopNode setAttribute ref $xp
    }
    $OP removeAllAttributes constellation
    $OP removeAllAttributes name
    set doc [$OP ownerDocument]
    return [$OP asXML]
    $doc delete
}






# In the entire opetope document, remove all $attr attributes:
proc ::dom::domNode::removeAllAttributes { domNode attr } {
    set doc [$domNode ownerDocument]
    foreach N [$doc selectNodes descendant-or-self::*\[@$attr\]] {
	$N removeAttribute $attr
    }
}
# One important use of this proc is to remove all those weight attributes!
# But in fact we also use it internally to remove name attributes before
# comparing...


# This proc applies a name and ref substitution as specified by $map.  
# The scope of the substitution is the whole owner opetope of $domNode.
# We do not check the validity of the opetope, nor do we care whether
# the map is a bijection, nor do we check the resulting opetope...
# The caller is supposed to take care of all this.
proc ::dom::domNode::applyMap { domNode map } {
    array set M $map
    foreach N [$domNode selectNodes //*] {
	set name [$N @name]
	if { [::info exists M($name)] } {    
	    $N setAttribute name $M($name)
	}
	if { [$N hasAttribute ref] } {
	    set ref [$N @ref]
	    if { [::info exists M($ref)] } {    
		$N setAttribute ref $M($ref)
	    }
	}
    }
}



# --------------------------------------------------------------------
# SECTION 5  --  Gluing
# --------------------------------------------------------------------

# Given an opetope node, a specified source of it (name of dot in composition
# tree), and another opetope whose target matches that source, glue the second
# opetope on top of the first.  In other words, substitute the second into the
# specified dot of the first.  The first opetope is modified by this operation:
# it becomes the result of the operation.  The second is not modified.
proc ::dom::domNode::glueOnto { opetope sourceFace topOpetope } {
    $opetope checkOpetope
    $topOpetope checkOpetope
    set theSpecificSource [$opetope Source $sourceFace]
    set topsTarget [$topOpetope Target]
    if { ![areIsomorphic $topsTarget $theSpecificSource bijMap] } {
	error "Could not verify compatibility condition.\n(Could not \
	  find an isomorphism between the specified source of the bottom opetope\
	  and the target of the top opetope.)"
    }
    # Clone the topOpetope and substitute names to make its target equal to
    # the receiving source.  The problem is that we cannot just apply the
    # bijection map we found, because that only referred to the target
    # (i.e. all constellations but the last), and there is a risk that some
    # of the new labels we will introduce wiil interfere with the names
    # precent in top constellation.
    set top [$topOpetope cloneOpetopeNode]
    set replacementConst [$top lastChild]
    set bottomNames [list]
    foreach D [$opetope selectNodes descendant::dot] {
	lappend bottomNames [$D @name]
    }
    set max 0
    foreach upperD [$replacementConst selectNodes descendant::dot] {
	if { [lcontain $bottomNames [$upperD @name]] } {
	    # The top opetope uses a name which clashes with some name used in
	    # the bottom opetope (but not seen in the specific soruce).
	    # We need to invent a new name.  We take the smallest natural
	    # number not already in use as a name anywhere.  So first find the
	    # biggest natural number in use:
	    foreach thisName $bottomNames {
		if { [string is integer -strict $thisName] && ($thisName > $max) } {
		    set max $thisName
		}
	    }
	    foreach D [$top selectNodes descendant::dot] {
		set thisName [$D @name]
		if { [string is integer -strict $thisName] && ($thisName > $max) } {
		    set max $thisName
		}
	    }
	    lappend bijMap [$upperD @name] [expr {$max+1}]
	}
    }
    # Now we can finally apply the bijection map
    $top applyMap $bijMap
    $top checkOpetope
    # At this point, the names have all been compatibilised, and there are no
    # name clashes either, so now we can just substitute the tree from top into
    # the dot in bottom
    set ct [$opetope lastChild]
    set receivingNode [$ct getNodeWithName $sourceFace]
    set replacementRoot [$replacementConst firstChild]
    $receivingNode replaceByTree $replacementRoot
    $opetope setAttribute name "[$opetope @name]+$sourceFace+[$topOpetope @name]"
    $opetope checkOpetope
    return $opetope
}


# This proc substitutes a dot by a tree of the 'same shape', it the total
# bouquet of the replacement tree is equal to the local bouquet of the dot
# we substitute into.  We only work in this level, so before trying to do
# this, it should be checked that the decorations (lower cosntellations)
# really match!  In fact, in this proc, 'same shape' means not only a
# bijection, but indeed that the very same names are used for leaves of
# the replacement tree and child nodes of the dot we substitute into.
# To calling proc [glueOnto] checks isomorphisms carefully and arrange
# the names to match exactly.
proc ::dom::domNode::replaceByTree { receivingNode replacementTree } {
    # First check compatibility:
    if { [$receivingNode nodeName] ne "dot" } {
	error
    }
    set childIDs [list]
    foreach child [$receivingNode childNodes] {
	lappend childIDs [$child @name]
    }
    set childIDs [lsort $childIDs]
    set leafIDs [list]
    foreach leaf [$replacementTree selectNodes descendant-or-self::leaf] {
	lappend leafIDs [$leaf @name]
    }
    set leafIDs [lsort $leafIDs]
    if { $childIDs ne $leafIDs } {
	error "Couldn't compatibilise node with replacement tree"
    }
    # For each child of the receiving node, remove this child and
    # replace it into the leaves of the replacement tree.
    foreach child [$receivingNode childNodes] {
	set id [$child @name]
	$receivingNode removeChild $child
	set oldLeaf [$replacementTree find name $id]
	set parent [$oldLeaf parentNode]
	$parent replaceChild $child $oldLeaf
	$oldLeaf delete
    }
    set replacementParent [$replacementTree parentNode]
    # Finally set the big tree into the receiving node:
    set receivingParent [$receivingNode parentNode]
    $receivingParent replaceChild $replacementTree $receivingNode
    
#     return [$receivingParent asXML]
}



# --------------------------------------------------------------------
# SECTION 6  --  Reading and writing files
# --------------------------------------------------------------------

# Read and parse and XML file.
# Return a handle to the opetope node of the in-memory xml document.
proc readOpetope { xmlFile } {
    set fd [open $xmlFile r]
    set xml [read $fd]
    close $fd
    set doc [dom parse -baseurl $::baseURI $xml]
    set opetopeNode [$doc documentElement]
    if { [catch {$opetopeNode checkOpetope} err] } {
	$doc delete
	error "Not a valid opetope file: $err"
    }
    return $opetopeNode
}
# This opetope node can then be used and modified, for example by
# performing sphere operations on the top constellation, and writing the
# resulting opetopes back to new xml files.
# Example:
#   set OP [readOpetope example.xml]
#   set topConst [$OP lastChild]
#   set someDot [lindex [$upper selectNodes descendant-or-self::dot] 0]
#   set someDotName [$someDot @name]
#   $topconst selectSaturatedSubtree $someDotName
#   # This has now modified $OP, since $topconst is part of $OP
#   # Write this new opetope to another file:
#   $OP writeOpetope anotherExample.xml
#   set doc [$OP ownerDocument]
#   $doc delete


# Given an opetope node, write it to an xml file.
proc ::dom::domNode::writeOpetope { opetopeNode fileName } {
    return [::writeOpetope $opetopeNode $fileName]
}
proc writeOpetope { opetopeNode fileName } {
    if { [$opetopeNode nodeName] ne "opetope" } {
	error "$opetopeNode (named \"[$opetopeNode @name]\" not an opetope node"
    }
    if { [catch {$opetopeNode checkOpetope} err] } {
	error "$opetopeNode (named \"[$opetopeNode @name]\" not a valid opetope\n$err"
    }
    set fileName [file normalize $fileName]
    set dirName [file dirname $fileName]
    if { ![file exists $dirName] } {
	file mkdir $dirName
    }
    set doc [$opetopeNode ownerDocument]
    set fd [open $fileName w]
    puts $fd {<?xml version="1.0" encoding="UTF-8"?>}
    $doc asXML -channel $fd -doctypeDeclaration 1
    close $fd
    if { [namespace exists ::alpha] } {
	setFileInfo $fileName type TEXT
	setFileInfo $fileName creator ALFA
    }
    return $fileName    
}


# --------------------------------------------------------------------
# SECTION 7  --  File-level procs
# --------------------------------------------------------------------
# Procedures applying to opetope xml files.  These are designed to be
# used interactively.  They never modify the xml files, and will write
# any output opetopes to new xml files.  


proc checkOpetopeFile { xmlFile } {
    # By reading the file we already check it:
    set opetopeNode [readOpetope $xmlFile]
    # Just clean up again:
    set doc [$opetopeNode ownerDocument]
    $doc delete
    # Since no error occurred, we have
    return OK
}

# Return the names of the sources (i.e. dots in the last
# constellation (the composition tree)):
proc listSources { xmlFile } {
    set opetopeNode [readOpetope $xmlFile]
    set upper [$opetopeNode lastChild]
    set dotList [$upper selectNodes descendant-or-self::dot]
    set dotNames [list]
    foreach dot $dotList {
	lappend dotNames [$dot @name]
    }
    set doc [$opetopeNode ownerDocument]
    $doc delete
    return $dotNames
}


# Remark:  
#    $upper selectNodes descendant-or-self::dot
# produces a list of domNodes.  We would like to have a list of name attributes.
# Doing instead
#    $upper selectNodes descendant-or-self::dot/@name
# will produce a list like {{name 23} {name 24} {name 35}}



proc areIsomorphic { xmlFile0 xmlFile1 } {
    set op0 [readOpetope $xmlFile0]
    set op1 [readOpetope $xmlFile1]
    if { [::dom::domNode::areIsomorphic $op0 $op1 map] } {
	return "Opetopes isomorphic."
    } else {
        return "The required bijections could not be found (but might exist)."
    }
}


# Given an opetope xml file and a source specified by name (i.e. a dot in
# the top constellation (the composition tree), compute that source opetope
# and write it to a new opetope file.  By default the new file name will be
# invented automatically: If the original opetope is called foo, and the
# source 13, then the new opetope will be written to foo-13.xml.
# Another name can be specified by supplying an extra argument like this
#    computeOneSource xmlFile 13 -o newFile
# (In any case, the path to the written file is returned by the proc.)
proc computeOneSource { xmlFile sourceName args } {
    # The otional parameters is only to indicate output name:
    # if there are precisely two words and the first is "-o"
    # then the second will be taken as a filename to write the result to.
    if { [llength $args] == 2 } {
	if { [lvarpop args] eq "-o" } {
	    set newFileName [lvarpop args]
	}
    }
    # Read in the opetope:
    set opetopeNode [readOpetope $xmlFile]
    # Compute the source:
    set newOpetope [$opetopeNode Source $sourceName]
    # Harmonise names:
    if { [info exists newFileName] } {
	set newName [file root [file tail $newFileName]]
	$newOpetope setAttribute name $newName
    } else {
	set dir [file dirname $xmlFile]
	set newFileName [file join $dir [$newOpetope @name].xml]
    }
    # Write the result to a new file:
    writeOpetope $newOpetope $newFileName
    # Clean up
    set doc [$opetopeNode ownerDocument]
    $doc delete
    set doc [$newOpetope ownerDocument]
    $doc delete
    # and return the name of the written file (without the leading ./):
    regsub {^\./} $newFileName "" newFileName
    return $newFileName    
}


# Given an opetope xml file, compute the target and write it to a new
# opetope file.  By default the new file name will be invented
# automatically: if the original opetope is called foo, then the new
# opetope will be written to foo-%.xml.  Another name can be
# specified by supplying an extra argument like this
#    computeTarget xmlFile -o newFile
# (In any case, the path to the written file is returned by the proc.)
proc computeTarget { xmlFile args } {
    # The otional parameters is only to indicate output name:
    # if there are precisely two words and the first is "-o"
    # then the second will be taken as a filename to write the result to.
    if { [llength $args] == 2 } {
	if { [lvarpop args] eq "-o" } {
	    set newFileName [lvarpop args]
	}
    }
    # Read in the opetope:
    set opetopeNode [readOpetope $xmlFile]
    # Compute the target:
    set newOpetope [$opetopeNode Target]
    # Harmonise names:
    if { [info exists newFileName] } {
	set newName [file root [file tail $newFileName]]
	$newOpetope setAttribute name $newName
    } else {
	set dir [file dirname $xmlFile]
	set newFileName [file join $dir [$newOpetope @name].xml]
    }
    # Write the result to a new file:
    writeOpetope $newOpetope $newFileName
    # Clean up
    set doc [$opetopeNode ownerDocument]
    $doc delete
    set doc [$newOpetope ownerDocument]
    $doc delete
    # and return the name of the written file (without the leading ./):
    regsub {^\./} $newFileName "" newFileName
    return $newFileName    
}


proc computeAllFacets { xmlFile } {
    computeTarget $xmlFile
    foreach source [listSources $xmlFile] {
	computeOneSource $xmlFile $source
    }
}
# # Instead of reading in and parsing a new copy of the orginal for each facet,
# # keep the thing in memory:
# proc altComputeAllFacets { xmlFile } {
#     set opetopeNode [readOpetope $xmlFile]
#     set originalOpetopeName [$opetopeNode @name]
#     set facetDir [file join [file dirname $xmlFile] "${originalOpetopeName}.facets"]
#     # Target:
#     set targetNode [$opetopeNode Target]
#     set newFileName [file join $facetDir ${newOpetopeName}.xml]
#     writeOpetope $opetopeNode $newFileName
#     # Clean up
#     set doc [$targetNode ownerDocument]
#     $doc delete
#     # Sources:
#     foreach sourceName [$opetopeNode listSourceNames] {
# 	set sourceNode [$opetopeNode Source $sourceName]
# 	set newFileName [file join $facetDir ${newOpetopeName}.xml]
# 	writeOpetope $opetopeNode $newFileName
# 	# Clean up
# 	set doc [$sourceNode ownerDocument]
# 	$doc delete
#     }
#     # Clean up original:
#     set doc [$opetopeNode ownerDocument]
#     $doc delete
#     return $facetDir
# }


proc glueOnto { xmlFileBottom sourceName xmlFileTop args } {
    # The otional parameters is only to indicate output name:
    # if there are precisely two words and the first is "-o"
    # then the second will be taken as a filename to write the result to.
    if { [llength $args] == 2 } {
	if { [lvarpop args] eq "-o" } {
	    set newFileName [lvarpop args]
	}
    }
    set opetope [readOpetope $xmlFileBottom]
    set topOpetope [readOpetope $xmlFileTop]
    ::dom::domNode::glueOnto $opetope $sourceName $topOpetope
    # Harmonise names:
    if { [info exists newFileName] } {
	set newName [file root [file tail $newFileName]]
	$opetope setAttribute name $newName
    } else {
	set dir [file dirname $xmlFileBottom]
	set newFileName [file join $dir [$opetope @name].xml]
    }
    writeOpetope $opetope $newFileName
    # Clean up:
    set doc [$opetope ownerDocument]
    $doc delete
    set doc [$topOpetope ownerDocument]
    $doc delete
    # and return the name of the written file (without the leading ./):
    regsub {^\./} $newFileName "" newFileName
    return $newFileName    
}




# --------------------------------------------------------------------
# SECTION 8  --  Some auxiliary procs
# --------------------------------------------------------------------

# The command [lrepeat] of Tcl8.5 is quite stupid in that it doesn't allow
# the empty list.  Instead we use this:

proc lrepeatAllowingZero {count value args} {
    if { ![string is integer -strict $count] } {
	error "expected integer but got \"$count\""
    } elseif { $count < 1 } {
	return [list]
    }
    set values [linsert $args 0 $value]
    set result {}
    for {set i 0} {$i < $count} {incr i} {
	eval [list lappend result] $values
    }
    return $result
}


# Given a list of vectors of equal lenght, return their vector sum
# (component-wise sum).
proc sumWeights { L } {
    # L is a list of vectors of equal length
    set len [llength [lindex $L 0]]
    set res [lrepeatAllowingZero $len 0]
    foreach V $L {
	if { [llength $V] != $len } {
	    error "$V not of length $len"
	}
	set newRes [list]
	for {set i 0} {$i<$len} {incr i} {
	    lappend newRes [expr {[lindex $V $i] + [lindex $res $i]}]
	}
	set res $newRes
    }
    return $res
}


# A bijection map is a even-length list like {a 1 b 2 c 3} in which neither
# the sublist of odd elements nor the sublist of even elements has any
# repetition.
proc isBijectionMap { map } {
    if { [llength $map] % 2 } {
	return 0
    }
    set L0 [list]
    set L1 [list]
    foreach {x0 x1} $map {
	lappend L0 $x0
	lappend L1 $x1
    }
    if { [llength [lsort -unique $L0]] != [llength $L0] } {
	return 0
    }
    if { [llength [lsort -unique $L1]] != [llength $L1] } {
	return 0
    }
    return 1    
}

# Compute the inverse of a map (assuming it is a bijection map):
proc inverse { map } {
    set newMap [list]
    foreach {x0 x1} $map {
	lappend newMap $x1 $x0
    }
    return $newMap
}


# Copyright (c) 2004-2006, Joachim Kock.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#    * Redistributions of source code must retain the above copyright 
#      notice, this list of conditions and the following disclaimer.
# 	   
#    * Redistributions in binary form must reproduce the above copyright 
#      notice, this list of conditions and the following disclaimer in the 
#      documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
